;  Scrolling 3D 'Ringworld' Road Engine for ZX Sectrum
;  ---------------------------------------------------
;
;             by Jason "Joefish" Railton
;             --------------------------
;
;
;NOTES:
;-----
;
;48K model initial version
;
;Interrupts arranged around possible 128K memory paging.


ENTRY_POINT 	EQU 32768
TOP_PAGE        EQU 49152

MACRO PAGE_ALIGN
	defs (65536-$)AND 255
ENDM

MACRO WAIT_VBL
	ei
	halt
ENDM

DEFAULT_BORDER	EQU	0
DEBUG_BORDER_1	EQU 0

DEBUG_BORDER_2	EQU DEBUG_BORDER_1*2
DEBUG_BORDER_3	EQU DEBUG_BORDER_1*3
DEBUG_BORDER_4	EQU DEBUG_BORDER_1*4
DEBUG_BORDER_5	EQU DEBUG_BORDER_1*5
DEBUG_BORDER_6	EQU DEBUG_BORDER_1*6
DEBUG_BORDER_7	EQU DEBUG_BORDER_1*7



FIRE_JUMP_CUTOFF        EQU 10
NORMAL_TARGET_SPEED     EQU 768
SLOW_TARGET_SPEED       EQU 64
BOOST_TARGET_SPEED      EQU 1024

TIME_LOOPS_PER_SECOND   EQU 50



MAP_DATA_SIZE           EQU 1024

MAP_NAME_COLOUR_OFFSET   EQU 0
MAP_AUTHOR_COLOUR_OFFSET EQU 4

MAP_NAME_SIZE           EQU 11
MAP_NAME_OFFSET         EQU 8
MAP_AUTHOR_SIZE         EQU 17
MAP_AUTHOR_OFFSET       EQU 19

MAP_BOOST_OFFSET        EQU 36
MAP_LENGTH_OFFSET       EQU 37
MAP_TIME_OFFSET         EQU 39
MAP_PALETTE_OFFSET      EQU 42
MAP_PALETTE_SIZE        EQU 44

MAP_MAP_OFFSET          EQU 262
MAP_MAP_SIZE            EQU 762


RUN_IN_SIZE             EQU 10
OFF_ROAD_DELTA          EQU 15


PLAYER_ROW_OFFSET       EQU 238
BOTTOM_ROW_OFFSET       EQU 240


;Initial entry point
;-------------------
org ENTRY_POINT


;Initialise
;----------
	di
	
	ld sp,49152			;Top of the safe middle 16K

	
;Memory Organisation
;-------------------


;Configure interrupts
;--------------------
;Interrupts will go to 189*256+189 = 48573
;Just stick a RET at this address.  The interrupt only exists
;to be able to time things.
;Note that a simple RET (with no EI) will leave interrupts
;disabled on return.  They must be re-enabled before any use
;of HALT for timing.
	ld a,$c9		;ret
	ld (48573),a

;Configure interrupt vector table,
;set 48640..48896 (257 bytes) to 189
;do this last so most of this code can be over-written.
	ld hl,48640		;190*256=48640
  	ld de,48641
	ld (hl),189		;write once
	ld bc,256       ;make 256 copies
	ldir

;This leaves 255 bytes free from 48897 to 49152, the upper part of the middle 16K of RAM.
;And also 66 bytes at 48574, between the RET and the table.
;The memory below the RET at 487573 is then all free.	
	
;Install interrupts:
	ld a,190		;high byte of address table
	ld i,a
	im 2


;Prep Screen:
;------------

    ld hl,pixel_map
    ld de,16384
    ld bc,6400
    ldir
    
    ld hl,22528+256
    push hl
    ld de,22528+256+1
    ld bc,31
    ld (hl),7
    ldir
    ld (hl),0
    pop hl
    ld bc,512-32
    ldir
  
     
;Prep Multicolour:
;-----------------

    ld hl,colour_map_a    
    ld de,colour_map_a+1
    ld bc,1967
    ld (hl),7
    ldir
    
    
;Init:    
;-----
    
    xor a
    ld (road_row),a
    ld (road_row_was),a
    ld (flag_road_new),a
    
    ld bc,0
    ld (road_pos),bc
    ld (road_speed),bc
    ld (road_count),bc
    ld bc,NORMAL_TARGET_SPEED
    ld (tgt_speed),bc
    
    ld iy,road_data
    ld (road_base),iy
    
    ld a,(iy+MAP_TIME_OFFSET+0)
    ld (time+0),a
    ld a,(iy+MAP_TIME_OFFSET+1)
    ld (time+1),a
    ld a,(iy+MAP_TIME_OFFSET+2)
    ld (time+2),a
    
    ld a,(iy+MAP_LENGTH_OFFSET+0)
    ld l,a
    ld a,(iy+MAP_LENGTH_OFFSET+1)
    ld h,a
    dec hl
    ld (road_length),hl
    
    ld a,(iy+MAP_BOOST_OFFSET)
    and 7
    rlca
    rlca
    rlca
    ld (boost_colour),a
    
    ld bc,MAP_PALETTE_OFFSET
    add iy,bc
    ld (palette_ptr),iy
    
    ld bc,MAP_MAP_OFFSET-MAP_PALETTE_OFFSET
    add iy,bc
    ld (road_ptr),iy
    
    ld bc,(road_length)
    add iy,bc
    ld bc,RUN_IN_SIZE+3
    add iy,bc
    ld (end_road_ptr),iy
    
    ld a,2
    ld (x_slot),a
    
    ld a,14
    ld (x_tgt),a
    ld (x_sprite),a
    ld (x_erase),a
    ld a,7
    ld (y_sprite),a
    
    xor a
    ld (blink),a
    ld (control_right),a
    ld (control_left),a
    ld (control_fire),a
    ld (control_was_fire),a    
    ld (jump_seq),a
    ld (jump_sound),a
    
    ld (jump_dec_boost),a
    ld (boost_inc),a
    ld (boost_ting),a
    ld (boost_noise),a
    ld (flag_boost_noise),a
    ld (noise_index),a
    
    ld a,00100111b
    ld (time_colour),a
    
    ld a,255
    ld (boost_level),a
    
    ld a,4
    ld (flag_clock),a
    
    ld a,1
    ld (off_road),a
    
    
;Pre-load road:
;--------------

pre_load_road_loop
    ld a,(road_row)
    sub 16
    ld (road_row),a
    call do_road_new
    call do_road_resume
    ld a,(road_row)
    and a
    jp nz,pre_load_road_loop
    	   
        	
;Main loop:
;----------
	   
main_loop
main_loop_cycle_1
	call map_road_a         ; red border time		
	call input_controls
    call game_logic         ; magenta
    call player_controls    ; yellow
    call move_road          ; cyan
    call move_sprite        ; red
    call sound_fx           ; white
    call process_time       ; green
    call process_events     ; white
	WAIT_VBL
	call draw_sprite        ; blue
	call map_road_b         ; magenta
	call top_wait           ; cyan
	call multicolour        ; 8x1 blue/black (with sound), 8x2 black

main_loop_repeat		
    jp main_loop

        
;Variables:
;----------
    
    road_pos    dw 0
    
    road_row        db 0
    road_row_was    db 0
    flag_road_new   db 0
    
    road_base   dw 0
    road_count  dw 0
    road_length dw 0
    road_ptr    dw 0
    end_road_ptr dw 0
    palette_ptr dw 0
    
    blink       db 0
    off_road    db 1
    finished    db 0
    
    road_speed  dw 0
    tgt_speed   dw 0
    
    go_slow     db 0
    
    x_slot      db 2
    x_tgt       db 14

    x_dummy     db 0

    x_sprite    db 14,0    
    y_sprite    db  7,0   
    
    x_erase     db 14,0
    sp_erase    dw screen_table
    
    jump_seq    db 0
    jump_sound  db 0                                                         
    sound_lr    db 0
    
    control_right       db 0
    control_left        db 0
    control_fire        db 0
    control_was_fire    db 0
    
    out_pointer         dw sound_table_0  

    time_colour         db 00100111b
    time                db 2, 0, 0, TIME_LOOPS_PER_SECOND
    flag_clock          db 0
    
    boost_level         db 255
    jump_dec_boost      db 0  
    boost_inc           db 0
    
    boost_colour        db 0
    boost_ting          db 0
    boost_noise         db 0
    flag_boost_noise    db 0
    noise_index         db 0
    
    
;Game Logic:
;-----------

game_logic

	ld a,DEBUG_BORDER_3
	out (254),a

	xor a
	ld (go_slow),a
	
; Road detection
    ld a,(boost_colour)
    ld d,a
    
    ; Check not in start / end:
    ld a,(off_road)
    and a
    jp nz,no_road_contact

    ; Get colour under player:    
    ld a,(x_sprite)
    rlca
    ld hl,collision_table
    ld b,0
    ld c,a
    add hl,bc
    ld a,(hl)
    inc hl
    ld h,(hl)
    ld l,a
    ld a,(hl)
    and 56
    
    ; Jump tile (White)
    cp 56                   ;Check white jump square
    jp nz, road_not_white
    ld a,(jump_seq)
    cp JUMP_MIN
    jp nc,road_not_white
    ld a,JUMP_MAX
    ld (jump_seq),a    
road_not_white		

    ex af,af'
    ; Check not blinking:
    ld a,(blink)
    and a
    jp nz,no_road_contact
    
    ; Check not bouncing:
    ld a,(y_sprite)
    cp 7
    jp nz,no_road_contact
    ex af,af'

    ; Empty tile (black)
    and a
    jp nz,road_not_empty
    jp fall_down
road_not_empty    
        
    ; Boost tile
    cp d
    jp nz,road_not_boost
    ld bc,BOOST_TARGET_SPEED
    ld (tgt_speed),bc
    ld (road_speed),bc
    ld a,(boost_ting)
    and a
    jp nz,boost_skip_ting
    ld a,BOOST_TING_TIME+1
    ld (boost_ting),a
boost_skip_ting
road_not_boost

    ; Slow tile (Green)
    cp 32
    jp nz,road_not_slow
    ld bc,SLOW_TARGET_SPEED
    ld (tgt_speed),bc
    ld (road_speed),bc
	ld (go_slow),a
road_not_slow

no_road_contact
	
    ret
    
	
; Player controls:

player_controls
	ld a,DEBUG_BORDER_6
	out (254),a

    xor a
    ld (sound_lr),a
    ld (boost_noise),a

    ld a,(go_slow)
    and a
    jp nz,skip_controls
    
check_move_left	
	ld a,(control_left)
	and 7
	ld (control_left),a
	cp 1
	jp nz,no_move_left
move_left
    ld a,(x_slot)
    sub 1               ; Use SUB because DEC A doesn't trigger a CARRY flag
    jp c,no_move_left
    ld (x_slot),a       ; Move one slot left
	ld a,(x_tgt)
	ld (x_sprite),a     ; If still moving to previous slot, jump there instantly
    ld a,1
    ld (sound_lr),a
no_move_left
	
check_move_right	
	ld a,(control_right)
	and 7
	ld (control_right),a
	cp 1
	jp nz,no_move_right
move_right
    ld a,(x_slot)
    inc a
    cp 5
    jp z,no_move_right
    ld (x_slot),a       ; Move one slot right
	ld a,(x_tgt)
	ld (x_sprite),a     ; If still moving to previous slot, jump there instantly
    ld a,1
    ld (sound_lr),a
no_move_right
	
    ; No jumping / boosting on run-up:
    ld a,(off_road)
    and a
    jp nz,skip_controls
    
check_jump
    ld a,(jump_dec_boost)
    sub 1
    jp c,no_jump_dec_boost
    ld (jump_dec_boost),a
   	ld a,(boost_level)
   	and a
   	jp z,no_jump_dec_boost
   	sub 5
    jp nc,jump_boost_decced
    xor a
jump_boost_decced
    ld (boost_level),a
    xor a
    ld (boost_inc),a
no_jump_dec_boost

    ld a,(jump_seq)
    cp JUMP_MIN
    jp nc,move_no_jump
    
   	ld a,(boost_level)
    cp JUMP_BOOST_COST-4
    jp c,move_no_jump
    
    ld a,(control_was_fire)
    and a
    jp z,move_no_jump
    cp FIRE_JUMP_CUTOFF
    jp nc,move_no_jump
    ld a,JUMP_MAX
    ld (jump_seq),a
    ld a,JUMP_BOOST_COST/5
    ld (jump_dec_boost),a
move_no_jump
move_jump_done        

check_boost
    ld bc,NORMAL_TARGET_SPEED
    ld (tgt_speed),bc
    ld a,(control_fire)
    cp FIRE_JUMP_CUTOFF
    jp c,check_boost_done
    xor a
    ld (boost_inc),a
	ld a,(boost_level)
	sub 1
	jp c,check_boost_done
	ld (boost_level),a
	
    ld hl,(road_speed)
    ld bc,4
    add hl,bc
    ld (road_speed),hl
do_boost
    ld a,1
    ld (boost_noise),a
    ld bc,BOOST_TARGET_SPEED
    ld (tgt_speed),bc
check_boost_done

    ld a,(jump_seq)
    and a
    jp nz,no_boost_inc
    ld a,(boost_inc)
    inc a
    and 3
    ld (boost_inc),a
    jp nz,no_boost_inc
    ld a,(boost_level)
    inc a
    jp z,no_boost_inc
    ld (boost_level),a
no_boost_inc  


skip_controls    
    ld a,(jump_seq)
    sub 1               ; Because DEC A doesn't Cy
    jp c,jump_seq_done
    ld (jump_seq),a
jump_seq_done
    ld a,(boost_ting)
    sub 1               ; Because DEC A doesn't Cy
    jp c,boost_ting_done
    ld (boost_ting),a
boost_ting_done

    ret
    

; Move road:
move_road
	ld a,DEBUG_BORDER_5
	out (254),a

    ld hl,(tgt_speed)
    ld d,h
    ld e,l
    ld bc,(road_speed)
    and a               ; Clear Cy
    sbc hl,bc           ; Like a CP. NZ if equal, Cy if BC>HL
	jp z,road_speed_done
	jp c,road_speed_dec
road_speed_inc
	inc bc
	inc bc
	inc bc
	inc bc              ; Cy not affected
	jp road_speed_done
	
road_speed_dec
	dec bc
	dec bc
	dec bc
	dec bc

	; Repeat test, so slow down faster than speed up:
	ex de,hl
    sbc hl,bc
	jp nc,road_speed_done
	dec bc
	dec bc
	dec bc
	dec bc
road_speed_done
	ld (road_speed),bc
	
	ld hl,(road_pos)
	  and a             ; Clear Cy
	  sbc hl,bc
	ld (road_pos),hl

    ld a,(road_row)
    ld (road_row_was),a
	ld c,a

	ld a,h              ; High byte of road pos
    and $F0             ; Row indicator in 16-row 256b cache
    ld (road_row),a
    cp c
    jp z,road_row_same
    ld a,2
    ld (flag_road_new),a
road_row_same
       	
    ret
    

move_sprite
	ld a,DEBUG_BORDER_2
	out (254),a		
	
	; Convert intended position from slot 0..4 to target X:
	ld a,(x_slot)
	ld b,a
	rlca
	rlca
	add a,b
	add a,4
	ld (x_tgt),a

		
	; Move towards target X:
	
	ld a,(x_tgt)
	ld b,a
	ld a,(x_sprite)
	cp b            ; NZ if equal, Cy if B>A
	jp z,sprite_x_move_done
	inc a           ; Cy not affected
	jp c,sprite_x_move_done
	dec a
	dec a
sprite_x_move_done

	ld (x_sprite),a

	
	; Jump sequence in Y:

sprite_y_move	
    ld a,(blink)
    sub 1
    jp c,not_blink_out
    
    ld (blink),a
    and 8
    jp z,not_blink_out
    ld a,24
    ld (y_sprite),a
    jp sprite_y_done
    
not_blink_out
	ld a,(jump_seq)
	ld c,a
	ld b,0
    ld hl,jump_table
    add hl,bc
    ld a,(hl)
    ld b,a
    and $0F
    ld (y_sprite),a
    xor b
    ld (jump_sound),a   ; top 4 bits of jump table = sound FX 0,1,2
sprite_y_done    
    
    ; Animate sprite:
    
sprite_animation
    ld a,(road_pos+1)
    rrca
    and 12
    ld b,0
    ld c,a
    ld hl,sprite_table
    add hl,bc
    ld e,(hl)
    inc hl
    ld d,(hl)
    inc hl
    ld (draw_sprite_pixels+1),de
    ld e,(hl)
    inc hl
    ld d,(hl)
    inc hl
    ld (draw_sprite_attrs+1),de
		
	ld a,(boost_level)
	srl a
	inc a
	srl a
	inc a
	srl a       ; div 8, 256 => 32
	ld b,0
	ld c,a
	ld hl,gauge_colours
	add hl,bc
	ld a,(hl)
	ld hl,gauge_table-2
	add hl,bc
	add hl,bc
	ld (hl),a
	inc hl
	inc hl
	xor a
	ld (hl),a
			
    ret
    
    
   ; Sound effects, starting with jump: 
sound_fx
	ld a,DEBUG_BORDER_7
	out (254),a		
     
sound_fx_a
	;top 4 bits of jump table index into sound table
	ld a,(jump_sound)
	and $30
	jp z,sound_fx_b
	ld e,a
	ld d,0
	ld hl,sound_tables
	add hl,de   ; x16
	rl e        ; x32
	rl e        ; x64
	add hl,de   ; x80
	ld (out_pointer),hl
	jp sound_fx_done

sound_fx_b
    ld a,(sound_lr)
    and a
    jp z,sound_fx_c
    ld hl,sound_table_3
	ld (out_pointer),hl
	jp sound_fx_done	 
    
sound_fx_c
    ld a,(boost_ting)
    cp BOOST_TING_TIME
    jp nz,sound_fx_d
    ld hl,sound_table_4
	ld (out_pointer),hl
	jp sound_fx_done	 

sound_fx_d
    ld a,(boost_noise)
    and a
    jp z,sound_fx_e
    ld (flag_boost_noise),a
	jp sound_fx_done
		
sound_fx_e
    ld hl,sound_table_0
	ld (out_pointer),hl
	;jp sound_fx_done	
sound_fx_done
	
    ret
	

;Controls:
;---------

input_controls
    jp input_keyboard


input_keyboard
    ld bc,$FBFE     ; Keyboard IN address, keys Q..T 
    in a,(c)        ; Bits 0..4 = Keys Q..T
    or 11100000b
    rlca            ; Shift up 1
    ld d,a
    ld bc,$DFFE     ; Keyboard IN address, keys P..Y
    in a,(c)        ; Bits 0..4 = P..Y
    or 11100000b
    and d           ; Combine keysets, offset by 1 bit
    cpl             ; Invert. Keys pressed are now 1s not 0s
    
    ld hl,input_keyboard_lr_done    ; Return address
    
    ld d,a              ; Copy for later
    and 00010101b       ; Check 'Right' keys
    jp nz,input_right
    ld a,d              ; Retrieve value
    and 00101010b       ; Check 'Left' keys
    jp nz,input_left
    
    jp input_no_lr
        
input_keyboard_lr_done  ; Return address
    ld bc,$7EFE         ; Both bottom rows
    in a,(c)
    cpl
    and 00011111b       ; Turn keys into 1s
    jp input_fire       ; Exit through a common handler
    
    
input_kempston
    ret
    
    
input_sinclair1
    ld bc,$EFFE     ; Keyboard IN address, keys 0..6
    in a,(c)        ; Bits 0..4 = Keys 0..6
    cpl             ; Invert. Keys pressed are now 1s not 0s
    ld d,a          ; For later
    
    ld hl,input_sinclair_lr_done  ; Return address

    bit 3,a
    jp nz,input_right
    bit 4,a
    jp nz,input_left
    
    jp input_no_lr
        
input_sinclair_lr_done  ; Return address
    ld a,d
    and 00000001b       ; Isolate FIRE button
    jp input_fire       ; Exit through a common handler


; Don't use D in these routines as it holds the input byte:        
    
input_right
    ld a,(control_right)
    inc a
    ld (control_right),a  
    xor a
    ld (control_left),a
    jp (hl)
    
    
input_left
    ld a,(control_left)
    inc a
    ld (control_left),a
    xor a
    ld (control_right),a
    jp (hl)

    
input_no_lr
    xor a
    ld (control_right),a
    ld (control_left),a
    jp (hl)

    
input_fire
    and a
    jp z,input_no_fire
    ld a,(control_fire)
    inc a
    ret z
    ld (control_fire),a
    ret
    
input_no_fire
    ld a,(control_fire)
    ld (control_was_fire),a
    out (254),a
    xor a
    ld (control_fire),a
    ret

    
;Process Time
;------------    
    
process_time
	ld a,DEBUG_BORDER_4
	out (254),a	
	
	ld a,(off_road)
	and a
	ret nz
	
	ld a,(finished)
	and a
	ret nz
	
dec_clock
	ld a,(time+3)
	sub 1
	ld (time+3),a
	jp nc,time_ok
	ld a,TIME_LOOPS_PER_SECOND
	ld (time+3),a
	ld a,24
	out (254),a
	
	ld a,4
	ld (flag_clock),a
	
	ld a,(time+2)
	sub 1
	ld (time+2),a
	jp nc,time_ok
	ld a,9
	ld (time+2),a
	
	ld a,(time+1)
	sub 1
	ld (time+1),a
	jp nc,time_ok
	ld a,5
	ld (time+1),a
	
	ld a,(time+0)
	sub 1
	ld (time+0),a
	jp nc,time_ok
	
	xor a
	ld (time+0),a
	ld (time+1),a
	ld (time+2),a
	jp time_zup	
time_ok

    ret

    
;Process Events
;------------    
process_events
	ld a,DEBUG_BORDER_7
	out (254),a
	
	ld a,(flag_road_new)
	and a
	jp nz,do_road_new
	
    ld a,(flag_clock)
    and a
    jp nz,render_clock

    ld a,(flag_boost_noise)
    and a
    jp nz,do_boost_noise

	ld a,DEFAULT_BORDER
	out (254),a
	
    ret

    
do_boost_noise
    xor a
    ld (flag_boost_noise),a
    
    ld a,(noise_index)
    inc a
    ld (noise_index),a
    ld h,5
    ld l,a
    ld a,(hl)
    and 63
    ld b,0
    ld c,a
    ld hl,sound_table_n
    add hl,bc
	ld (out_pointer),hl
	
	ld a,DEFAULT_BORDER
	out (254),a
	
    ret

            
do_road_new
    dec a
	ld (flag_road_new),a

    jp z,do_road_resume

    ld hl,(road_count)
    inc hl
    ld (road_count),hl
    
    ld bc,RUN_IN_SIZE+OFF_ROAD_DELTA
    xor a
    sbc hl,bc
    jp c,road_in_start
    ld (off_road),a

road_not_in_start    
    ld bc,(road_length)	
    xor a
    sbc hl,bc
    jp c,road_not_end
    ld a,1
    ld (off_road),a
    ld (finished),a
    ld a,4
    ld (flag_clock),a
    ld bc,0
    ld (tgt_speed),bc
        
road_in_start
road_not_end
    ld hl,(road_ptr)
    inc hl
    ld d,h
    ld e,l
    
    ld bc,(end_road_ptr)
    xor a
    sbc hl,bc
    jp c,road_not_data_end
    dec de
    dec de
   
road_not_data_end
    ex de,hl        
    ld (road_ptr),hl
    
    ld a,(hl)
    ld hl,(palette_ptr)
    ld b,0
    ld c,a
    add hl,bc
    
    ld d,road_map/256
    ld a,(road_row)
    ld e,a

    ld ixl,4
    ld a,(hl)
    and 64
    jp z,trans_row
    ld ixl,0            ; Bright rows don't show through stars
trans_row    

    ; Make the colour list row of 15 bytes:
    
    ld b,(hl)
    inc hl          ; Colour 1 in B

    ld a,b
    and 64+7        ; INK 1, PAPER black
    ld (de),a
    inc de

    ld a,b          ; INK 1, PAPER 1
    or ixl
    ld (de),a
    and 56
    jp nz,i1p1_ok
    ld a,7
    ld (de),a
i1p1_ok
    inc de
        
    ld c,(hl)
    inc hl          ; Colour 2 in C
        
    ld a,b
    xor c           ; Mess up A. XOR C a second time will restore
    and 7           ; Preserve messed up INK of Colour 1, 0 all else
    xor c           ; INK 1 restored, PAPER set to 2
    ld (de),a
    inc de

    ld a,b          ; Restore INK 1, PAPER 1
    xor c
    and 56
    xor c           ; INK 2, PAPER 1
    ld (de),a
    inc de
    
    ld a,c
    or ixl
    ld (de),a       ; INK 2, PAPER 2
    and 56
    jp nz,i2p2_ok
    ld a,7
    ld (de),a
i2p2_ok
    inc de

    ld (do_road_resume+1),hl
    ld (do_road_resume+4),de
    ld (do_road_resume+7),bc
    ld (do_road_resume+11),ix
    
	ld a,DEFAULT_BORDER
	out (254),a
	
    ret

        
do_road_resume
    ld hl,0000
    ld de,0000
    ld bc,0000    
    ld ix,0000
            
    ld b,(hl)
    inc hl          ; Colour 3 in B

    ld a,c  
    xor b
    and 56
    xor b           ; INK 2, PAPER 3
    ld (de),a
    inc de

    ld a,c          ; Restore INK 2, PAPER 2
    xor b
    and 7
    xor b           ; INK 3, PAPER 2
    ld (de),a
    inc de
    
    ld a,b
    or ixl
    ld (de),a       ; INK 3, PAPER 3
    and 56
    jp nz,i3p3_ok
    ld a,7
    ld (de),a
i3p3_ok
    inc de
    
    ld c,(hl)       ; Colour 4 in C
    inc hl
    
    ld a,b
    xor c
    and 7
    xor c           ; INK 3, PAPER 4
    ld (de),a
    inc de

    ld a,b          ; Restore INK 3, PAPER 3
    xor c
    and 56
    xor c           ; INK 4, PAPER 3
    ld (de),a
    inc de
    
    ld a,c
    or ixl
    ld (de),a       ; INK 4, PAPER 4
    and 56
    jp nz,i4p4_ok
    ld a,7
    ld (de),a
i4p4_ok
    inc de
        
    ld b,(hl)       ; Colour 5 in B
    
    ld a,c  
    xor b
    and 56
    xor b           ; INK 5, PAPER 4
    ld (de),a
    inc de

    ld a,c          ; Restore INK 4, PAPER 4
    xor b
    and 7
    xor b           ; INK 4, PAPER 5
    ld (de),a
    inc de
    
    ld a,b         
    or ixl
    ld (de),a       ; INK 5, PAPER 5
    and 56
    jp nz,i5p5_ok
    ld a,7
    ld (de),a
i5p5_ok
    inc de
    
    ld a,b
    and 64+7        ; INK 5, PAPER black
    ld (de),a
       
	ld a,DEFAULT_BORDER
	out (254),a
	
    ret
    
    
render_clock
    ld a,(flag_clock)
    dec a
    ld (flag_clock),a
    jp z,render_clock_sec
    dec a
    jp z,render_clock_ten
    dec a
    jp z,render_clock_min

render_clock_colour
    ld a,(finished)
    and a
    ld a,7*8+7
    jp nz,clock_colour_done
    
    ld a,(time+0)
    and a
    ld a,4*8+7
    jp nz,clock_colour_done
    
    ld a,(time+1)
    ld b,a
    cp 3
    ld a,4*8+7
    jp nc,clock_colour_done
    ld a,b
    and a
    ld a,6*8+7
    jp nz,clock_colour_done
    
    ld a,(time+2)
    cp 6
    ld a,6*8+7
    jp nc,clock_colour_done
    
    ld a,2*8+7
clock_colour_done
    ld (time_colour),a
    ld (22528+(12*32)+9),a
    ld (22528+(14*32)+9),a
    
  	ld a,DEFAULT_BORDER
	out (254),a
    ret
        
render_clock_sec    
    ld a,(time+2)
    ld de,22528+(11*32)+27
    jp display_time_digit
    
render_clock_ten    
    ld a,(time+1)
    ld de,22528+(11*32)+23
    jp display_time_digit

render_clock_min    
    ld a,(time+0)
    ld de,22528+(11*32)+5
    jp display_time_digit
    
    
display_time_digit  
    ld hl,time_table
    ld b,0
    ld c,a
    add hl,bc
    add hl,bc
    ld a,(hl)
    inc hl
    ld h,(hl)
    ld l,a
    
    ld a,(time_colour)
    ld c,a
    ld b,7
    
    jp(hl)

    include "rtp01_12_time.asm"    


;Fall Down    
;---------
    
fall_down
    pop bc      ; Clear return stack to game loop
    
    call fall_refresh_cycle
    
    xor a
    ld (sound_lr),a
    ld (boost_noise),a
    ld hl,sound_table_0
	ld (out_pointer),hl
    
    ld a,(road_pos+1)
    ld b,a
    and $0F
    cp 2
    jp nc,fall_here
    ld a,b
    dec a
    dec a
    ld (road_pos+1),a
fall_here
 
    ld a,(x_sprite)
    ld hl,drop_table
    ld b,0
    ld c,a
    add hl,bc
    ld a,(hl)
    ld (x_sprite),a
    
    ld a,8
    ld (fall_drop_count+1),a
fall_drop
    ld (y_sprite),a
    
    ld a,24
    out (254),a
    
    call fall_refresh_cycle

fall_drop_count
    ld a,8
    inc a
    inc a
    ld (fall_drop_count+1),a
    cp 26            	
    jp nz,fall_drop    

    ld a,100
    ld (fall_wait_count+1),a
fall_wait
    call fall_refresh_cycle	
fall_wait_count
    ld a,100
    dec a
    ld (fall_wait_count+1),a
    jp nz, fall_wait    

fall_resume
    xor a
    ld (jump_seq),a
    ld (control_right),a
    ld (control_left),a
    ld (control_fire),a
    ld (control_was_fire),a    
    ld (jump_seq),a
    ld (jump_sound),a
    
    ld (jump_dec_boost),a
    ld (boost_inc),a
    ld (boost_ting),a
    ld (boost_noise),a

    ld a,2
    ld (x_slot),a
    
    ld a,14
    ld (x_tgt),a
    ld (x_sprite),a
    
    ld a,7
    ld (y_sprite),a
    
    ld bc,NORMAL_TARGET_SPEED
    ld (tgt_speed),bc
    ld bc,0
    ld (road_speed),bc
    
    ld a,239
    ld (blink),a

	WAIT_VBL
	call draw_sprite        ; blue
	call map_road_b         ; magenta
	call top_wait           ; cyan
	call multicolour        ; 8x1 blue/black (with sound), 8x2 black
    jp main_loop
    
    
fall_refresh_cycle
	WAIT_VBL
	call draw_sprite        ; blue
	call map_road_b         ; magenta
	call top_wait           ; cyan
	call multicolour        ; 8x1 blue/black (with sound), 8x2 black
	call map_road_a         ; red border time		
    call process_time       ; green
    call process_events     ; white
	
    ret

    
;Time's Up
;---------

time_zup
    pop bc      ; Clear return stack to game loop

time_zup_loop
    call timezup_refresh_cycle	
    jp time_zup_loop    

    
timezup_refresh_cycle
	WAIT_VBL
	call draw_sprite        ; blue
	call map_road_b         ; magenta
	call top_wait           ; cyan
	call multicolour        ; 8x1 blue/black (with sound), 8x2 black
	call map_road_a         ; red border time		
	
    ret
    
            
;Top Wait
;--------
top_wait
    
	ld a,DEBUG_BORDER_5
	out (254),a
	
	;Pause for the top border:
	ld bc,8
top_wait_delay
    nop                         ;04t => 004t
    nop                         ;04t => 008t
	dec bc						;06t => 014t
	ld a,b						;04t => 018t
	or c						;04t => 022t
	jp nz,top_wait_delay	    ;10t => 032t per loop, 7 loops per 224t line
	
	nop
	nop
	nop
;	nop
;	nop
;	nop
;	nop
		
		
	ld a,DEFAULT_BORDER
	out (254),a
	
    ret
    
    
;Multicolour Macros
;------------------

MACRO POP_12B
    pop de
    pop bc
    pop af
    exx
    pop bc
    pop de
    pop hl
ENDM

MACRO PUSH_12B
    push hl
    push de
    push bc
    exx
    push af
    push bc
    push de
ENDM
    
MACRO POP_16BD
    pop de
    pop de
    pop de
    pop de
    exx
    exx
    pop de
    pop de
    pop de
    pop de
ENDM

MACRO PUSH_16BD
    push de
    push de
    push de
    push de
    exx
    exx
    push de
    push de
    push de
    push de     
ENDM
    

;Multicolour Processor
;---------------------

multicolour
    ld (multi_sp+1),sp

    
    ; Bottom corners of road normal attributes to match last 8x4 multicolour section:
    
    ld a,(colour_map_d+736-31)
    ld (22528+768-31),a
    ld a,(colour_map_d+736-3)
    ld (22528+768-3),a
    
    ; Sound effect pointer:
    
    ld hl,(out_pointer)
        
	; Dummy rows to stabilise multicolour:
	
    REPT 4,sync
        ld sp,white_row
        POP_16BD
       
        ld sp,22528+(9*32)-8
        PUSH_16BD
    ENDM

        
	; 8 Rows of alternate white INK / multicolour effect:
	
    REPT 4,multicol_a_rpt
        ld sp,white_row+2
        POP_12B
        
        ld sp,22528+(9*32)-10
        PUSH_12B
        
        ld a,(22528)
        ld (22528),a
        
        ld a,(hl)
        out (254),a
        inc hl    
        
        ld sp,colour_map_a+(multicol_a_rpt*16)+2
        POP_12B
        
        ld sp,22528+(9*32)-10
        PUSH_12B
                       
        ld a,(22528)
        ld (22528),a
        
        ld a,(hl)
        out (254),a
        inc hl    
        
    ENDM            
    
	;71 Rows of 16-wide 8x1 multicolour effect:
	
    REPT 71,multicol_b_rpt
        ld sp,colour_map_b+(multicol_b_rpt*16)+2
        POP_12B
        
        ld sp,22528+(((multicol_b_rpt/8)+10)*32)-10
        PUSH_12B          
        
        ld a,(end_gauge_table-9-multicol_b_rpt)
        ld (22528+(((multicol_b_rpt/8)+9)*32)+2),a
        
        ld a,(hl)
        out (254),a
        inc hl    

    ENDM
    
    
    ;Last of those will actually linger for 2 rows, giving the next
    ;stage time to start up:

    xor a
    out (254),a
    
    nop
    nop
    nop
    nop
    nop
    nop
    nop
    nop
    nop
    nop
    nop
    nop
    nop
    nop
    nop
    nop
    nop
    nop
    nop
    nop
    nop
    nop
    nop
    nop
    nop
    nop
    nop
    nop
    nop
    nop
    nop
    nop
            
    
    ;23 rows (46 pixel rows) of 8x2 multicolour, 27-wide:
    
    REPT 23,multicol_d_rpt
	    ld sp,colour_map_d+(multicol_d_rpt*32)+1
	    
        pop af
        pop de
        pop bc
        pop ix            
        exx   
        pop bc
        pop de
        pop hl
        pop iy
        exx   
        pop hl

        ld sp,22528+576+((multicol_d_rpt/4)*32)+9
	    ld (22528+576+((multicol_d_rpt/4)*32)+2),a
        push ix
        push bc
        push de
                         
	    ld sp,colour_map_d+(multicol_d_rpt*32)+19
        pop af
        pop bc                  
        pop de
            
        ld sp,22528+576+((multicol_d_rpt/4)*32)+25
        	        	  
	    push de 
	    push bc	
	    push af 
	    push hl
	    exx     
	    push iy 
	    push hl 
	    push de 
	    push bc
	        
		        
	    ld sp,colour_map_d+(multicol_d_rpt*32)+25
	    pop hl
	    ld (22528+576+((multicol_d_rpt/4)*32)+25),hl
	    pop hl
	    ld (22528+576+((multicol_d_rpt/4)*32)+27),hl
	        	
	    ;Timing adjust:
	        
	    nop
	    nop
	    nop
	    nop
	    
    ENDM
    
    
multi_sp
    ld sp,0000
    
    ret
    
    
;Road mapping:
;-------------

MACRO MAP_ROAD_HL_A
    ld a,(iy)
    inc iy
    add a,b
    and $F0
    ld l,a
ENDM

MACRO MAP_ROAD_HL_B
    ld a,(iy)
    inc iy
    add a,ixl
    and $F0
    ld l,a
ENDM

MACRO LDI_2
    ld a,(hl)
    inc l
    ld (de),a
    inc e
    ld (de),a
    inc e
ENDM

MACRO LDI_3
    ld a,(hl)
    inc l
    ld (de),a
    inc e
    ld (de),a
    inc e
    ld (de),a
    inc e
ENDM

MACRO LDI_4
    ld a,(hl)
    inc l
    ld (de),a
    inc e
    ld (de),a
    inc e
    ld (de),a
    inc e
    ld (de),a
    inc e
ENDM

MACRO LDI_5
    ld a,(hl)
    inc l
    ld (de),a
    inc e
    ld (de),a
    inc e
    ld (de),a
    inc e
    ld (de),a
    inc e
    ld (de),a
    inc e
ENDM
  
MACRO INK_ROAD
    ld a,(hl)
    and 7
    ld (de),a
    inc de
    inc l
    inc l
    inc l
    ld a,(hl)
    and 7
    ld (de),a
    inc de
    inc l
    inc l
    ld a,(hl)
    and 7
    ld (de),a
    inc de
    inc l
    inc l
    inc l
    inc l
    inc l
    inc l
    inc l
    ld a,(hl)
    and 7
    ld (de),a
    inc de
    inc l
    inc l
    ldi
ENDM
  

map_road_a    
    ld a,DEBUG_BORDER_2
    out (254),a
    
    ld a,(road_pos+1)
    ; Make sure we continue the same road pos for the upper half:    
    ld (map_road_b_road_pos+1),a
    ld b,a
    ld c,255                ; Protection against LDI affecting B. Only 155 LDIs in this section
    ld iy,road_table+1
    ld h,road_map/256

    include "rtp01_12_roada.asm"
    
    ld a,DEFAULT_BORDER
    out (254),a
    
    ret

    
map_road_b    
    ld a,DEBUG_BORDER_3
    out (254),a
    
map_road_b_road_pos
    ld a,0
    ld ixl,a
    ld iy,road_table+25
    ld h,road_map/256

    include "rtp01_12_roadb.asm"
    
    ld a,DEFAULT_BORDER
    out (254),a
    
    ret

    
;Draw Sprite
;-----------

draw_sprite
    ld (draw_sprite_restore_sp+1),sp
    
    
    ;Erase old sprite:
    
    ld a,DEBUG_BORDER_5
    out (254),a
       
    
     ; Erase old shadow pixels:
     
    ld hl,20704
    ld bc,(x_erase)
    ld (erase_sprite_add_x+1),bc
    add hl,bc
    ld d,h
    ld e,l
    ld bc,pixel_map-16384
    add hl,bc
    
    ldi
    inc hl
    inc e
    ld a,(hl)
    ld (de),a
    inc h
    inc d
    ldd
    dec hl
    dec e    
    ld a,(hl)
    ld (de),a
    inc h
    inc d
    ldi
    inc hl
    inc e
    ld a,(hl)
    ld (de),a
    inc h
    inc d
    ldd
    dec hl
    dec e    
    ld a,(hl)
    ld (de),a
       
    
    ; Erase old sprite pixels:
    
    ld sp,(sp_erase)
    ld a,14
    erase_sprite_loop
        ex af,af'
        
        pop hl
    erase_sprite_add_x
        ld bc,00
        add hl,bc
        ld d,h
        ld e,l
        ld bc,pixel_map-16384
        add hl,bc
        
        ldi
        ldi
        ld a,(hl)
        ld (de),a
        inc h
        inc d
        ldd
        ldd
        ld a,(hl)
        ld (de),a
    
        ex af,af'   ; retrieve loop counter
        dec a
    jp nz,erase_sprite_loop
        
    
    ; Draw sprite:
    
    ld a,DEBUG_BORDER_1
    out (254),a

        
    ; Draw shadow attributes:
    
    ld hl,colour_map_d+640
    ld bc,(x_sprite)
    ld (x_erase),bc
    ld (draw_sprite_add_x+1),bc
    add hl,bc
      ld a,(hl)
      and 01111000b
      ld (hl),a
      inc hl
    xor a
    ld (hl),a
    inc hl
      ld a,(hl)
      and 01111000b
      ld (hl),a
    ld bc,30
    add hl,bc
      ld a,(hl)
      and 01111000b
      ld (hl),a
      inc hl
    xor a
    ld (hl),a
    inc hl
      ld a,(hl)
      and 01111000b
      ld (hl),a
    
      
    ; Draw shadow pixels:
    
    ld hl,20704
    ld bc,(x_sprite)
    add hl,bc
    
    ld (hl),00000111b
    inc h
    ld (hl),01111111b
    inc h
    ld (hl),11111111b
    inc h
    ld (hl),00011111b
    ld bc,2-768
    add hl,bc
    ld (hl),11100000b
    inc h
    ld (hl),11111110b
    inc h
    ld (hl),11111111b
    inc h
    ld (hl),11111000b
    
    
    ; Attribute addresses:
    
    ld hl,colour_map_d
    ld bc,(y_sprite-1)      ; Gets y_sprite * 256
    srl b
    rr c
    srl b
    rr c
    srl b
    rr c                    ; divide by 8 gives y_sprite *32 
    add hl,bc
    ld bc,(x_sprite)
    add hl,bc
    ex de,hl
draw_sprite_attrs
    ld hl,sprite_0        
        
    exx         ; Save attribute address
    
    
    ; Pixel addresses:
    
    ld hl,screen_table
    ld bc,(y_sprite)
    add hl,bc
    add hl,bc
    ld (sp_erase),hl
    ld sp,hl
    
draw_sprite_pixels
    ld hl,sprite_0+84
    
    ld a,14    
    draw_sprite_loop
        ex af,af'   ; save loop counter
        exx         ; restore attribute address
        
        ; Sprite attribute data:
        ; attr 0
        ld b,(hl)   ; attribute mask but 1 bit high
        inc hl
        ld c,(hl)   ; colour
        inc hl
        xor a
        srl b       ; correct attribute mask
        sbc a,0     ; sign-extend spare bit into pixel mask
        ld iyh,a    ; save pixel mask 0
        
        ld a,(de)   ; get current attribute
        and b       ; apply mask
        xor c       ; apply colour
        ld (de),a   ; put attribute
        inc de
        
        ; attr 1
        ld b,(hl)   ; attribute mask but 1 bit high
        inc hl
        ld c,(hl)   ; colour
        inc hl
        srl b       ; correct attribute mask
        
        ld a,(de)   ; get current attribute
        and b       ; apply mask
        xor c       ; apply colour
        ld (de),a   ; put attribute
        inc de
        
        ; attr 2
        ld b,(hl)   ; attribute mask but 1 bit high
        inc hl
        ld c,(hl)   ; colour
        inc hl
        xor a
        srl b       ; correct attribute mask
        sbc a,0     ; sign-extend spare bit into pixel mask
        ld iyl,a    ; save pixel mask 2
        
        ld a,(de)   ; get current attribute
        and b       ; apply mask
        xor c       ; apply colour
        ld (de),a   ; put attribute
        
        ex de,hl
        ld bc,30
        add hl,bc
        ex de,hl    ; next attribute row
        
        ; Sprite pixel data:
        exx         ; restore pixel address
        
        ex de,hl
        pop hl      ; get screen address from table
    draw_sprite_add_x
        ld bc,00
        add hl,bc
        ex de,hl
        
        ld a,(de)   ; get screen byte
        and iyh     ; apply mask 0
        xor (hl)    ; apply pixels byte 0
        ld (de),a   ; put pixels
        inc hl
        inc e
        
        ldi         ; copy unmasked pixels byte 1
        
        ld a,(de)   ; get screen byte
        and iyl     ; apply mask 2
        xor (hl)    ; apply pixels byte 2
        ld (de),a   ; put pixels
        inc hl
        
        dec e
        dec e
        inc d
        
        ld a,(de)   ; get screen byte
        and iyh     ; apply mask 0
        xor (hl)    ; apply pixels byte 0
        ld (de),a   ; put pixels
        inc hl
        inc e
        
        ldi         ; copy unmasked pixels byte 1
        
        ld a,(de)   ; get screen byte
        and iyl     ; apply mask 2
        xor (hl)    ; apply pixels byte 2
        ld (de),a   ; put pixels
        inc hl    
        
        ex af,af'   ; retrieve loop counter
        dec a
    jp nz,draw_sprite_loop
    
    
    ld a,DEFAULT_BORDER
    out (254),a
    
draw_sprite_restore_sp
    ld sp,0000
    
    ret
    

;Sprites
;-------

sprite_table
    dw sprite_3, sprite_0123_attrs
    dw sprite_2, sprite_0123_attrs
    dw sprite_1, sprite_0123_attrs
    dw sprite_0, sprite_0123_attrs

sprite_0123_attrs
    db 11111111b,00000000b,  11110000b,00000000b,  11111111b,00000000b
    db 11110000b,00000000b,  00000000b,01010000b,  11110000b,00000000b
    db 00000000b,01010000b,  00000000b,01010000b,  00000000b,01010000b
    db 00000000b,01010000b,  00000000b,01010000b,  00000000b,00010000b

    db 00000000b,01010000b,  00000000b,01010000b,  00000000b,00010000b
    db 00000000b,01010000b,  00000000b,01010000b,  00000000b,00010000b
    db 00000000b,01010000b,  00000000b,01010000b,  00000000b,00010000b

    db 00000000b,01010000b,  00000000b,01010000b,  00000000b,00010000b
    db 00000000b,01010000b,  00000000b,01010000b,  00000000b,00010000b
    db 00000000b,01010000b,  00000000b,01010000b,  00000000b,00010000b

    db 00000000b,00010000b,  00000000b,00010000b,  00000000b,00010000b
    db 00000000b,00010000b,  00000000b,00010000b,  00000000b,00011000b
    db 11110000b,00000000b,  00000000b,00011000b,  11110000b,00000000b
    db 11111111b,00000000b,  11110000b,00000000b,  11111111b,00000000b
    
sprite_0
    db 00000000b,01111110b,00000000b
    db 00000000b,11111111b,00000000b
    db 00001111b,11000011b,11110000b
    db 01111111b,00011000b,11111110b
    db 11111000b,00111100b,00011111b
    db 11100001b,11001011b,10000111b
    db 11001111b,00000100b,11110011b
    db 11011000b,00001010b,00011011b

    db 11100000b,00000100b,00001111b
    db 11000010b,00000000b,00100011b
    db 11000000b,00000100b,00010111b
    db 11000000b,00000010b,10101011b
    db 11000011b,11100111b,11010111b
    db 11011100b,00011000b,00111111b
    
    db 11100000b,00000100b,01000111b
    db 11000010b,00001010b,00100011b
    db 11000000b,00000100b,01010111b
    db 11000000b,00000010b,00101011b
    db 11000000b,00000100b,01010111b
    db 11111100b,00000000b,00111111b

    db 11001111b,10000011b,11110111b
    db 11000000b,01100110b,00101011b
    db 11100010b,00011000b,01010111b
    db 11111000b,01000010b,00111111b
    db 01111111b,00010100b,11111110b
    db 00001111b,11001011b,11110000b
    db 00000000b,11111111b,00000000b
    db 00000000b,01111110b,00000000b

    
sprite_1
    db 00000000b,01111110b,00000000b
    db 00000000b,11111111b,00000000b
    db 00001111b,11000011b,11110000b
    db 01111111b,00111100b,11111110b
    db 11111011b,11100111b,11011111b
    db 11101110b,00001010b,01110111b
    db 11110000b,00000100b,00001111b
    db 11000010b,00000000b,00100011b
    
    db 11000000b,00000100b,01010111b
    db 11000000b,11100111b,00101011b
    db 11011111b,00011100b,11111111b
    db 11100000b,00001010b,00001111b
    db 11000000b,00000100b,01000111b
    db 11000010b,00000010b,00100011b
    
    db 11000000b,00000100b,01010111b
    db 11000000b,00000010b,00101011b
    db 11000000b,00000000b,01010111b
    db 11000000b,00000010b,10101011b
    db 11111111b,00000000b,01111111b
    db 11000001b,11100111b,11001011b
    
    db 11000000b,00011000b,00010111b
    db 11010010b,00000010b,10101011b
    db 11100100b,00000101b,01010111b
    db 11111010b,10000010b,10111111b
    db 01111111b,10110101b,11111110b
    db 00001111b,11001011b,11110000b
    db 00000000b,11111111b,00000000b
    db 00000000b,01111110b,00000000b
    
    
sprite_2
    db 00000000b,01111110b,00000000b
    db 00000000b,11111111b,00000000b
    db 00001111b,11000011b,11110000b
    db 01111111b,00100100b,11111110b
    db 11111010b,10000001b,01011111b
    db 11100110b,00000000b,10101111b
    db 11010000b,00000000b,00010111b
    db 11000000b,00011000b,00001011b
    
    db 11000001b,11100111b,10000111b
    db 11011111b,00001010b,11111011b
    db 11110000b,00000100b,00001111b
    db 11000000b,00000010b,10000011b
    db 11000010b,00000000b,01010011b
    db 11000000b,00000000b,00101011b
    
    db 11000000b,00000100b,01010111b
    db 11000000b,00000010b,00101011b
    db 11111110b,00000000b,01111111b
    db 11000011b,11100111b,11000011b
    db 11000000b,00011000b,00010111b
    db 11000010b,00000010b,00101011b
    
    db 11000000b,00000101b,01010111b
    db 11110000b,00001010b,10101111b
    db 11101110b,00000101b,11110111b
    db 11111011b,11100111b,10111111b
    db 01111111b,10111101b,11111110b
    db 00001111b,11001011b,11110000b
    db 00000000b,11111111b,00000000b
    db 00000000b,01111110b,00000000b

    
sprite_3
    db 00000000b,01111110b,00000000b
    db 00000000b,11111111b,00000000b
    db 00001111b,11000011b,11110000b
    db 01111111b,00000000b,11111110b
    db 11111000b,00011000b,00011111b
    db 11100000b,01111110b,00000111b
    db 11001111b,10000101b,11110011b
    db 11111000b,00001010b,00011111b
    
    db 11000000b,00000100b,00000011b
    db 11000000b,00000000b,00100011b
    db 11000010b,00000100b,01010111b
    db 11000000b,00000010b,10101011b
    db 11000000b,00000000b,01010111b
    db 11000000b,00000000b,00101011b
    
    db 11000011b,11100111b,11000111b
    db 11111100b,00011010b,00111111b
    db 11000000b,00000100b,01010111b
    db 11000010b,00001010b,00101011b
    db 11000000b,00000100b,01010111b
    db 11000000b,00000010b,00000011b
    
    db 11111100b,00000000b,00111111b
    db 11000111b,10000001b,11101011b
    db 11100000b,01100111b,01010111b
    db 11111010b,00011010b,10111111b
    db 01111111b,00010100b,11111110b
    db 00001111b,11001011b,11110000b
    db 00000000b,11111111b,00000000b
    db 00000000b,01111110b,00000000b
    
    
;Jump table:
;-------------

;4096 steps per square (16 x 256), so at 512 speed, 8 cycles per square.
;To jump over 2 squares, be airborne for 24 cycles.

;JUMP_MAX            EQU 50
;JUMP_MIN            EQU 29
;JUMP_BOOST_COST     EQU 86
;jump_table
;    db 7,7
;    db $27,6,6,5,5,5,6,6
;    db $07,$27,6,5,4,4,3,3, 3,3,3,3,4,4,5,$06
;    db $07,$27,5,4,3,2,2,1, 1,1,0,0,0,0,0,0, 1,1,1,2,2,3,4,$15

JUMP_MAX            EQU 34
JUMP_MIN            EQU 19
JUMP_BOOST_COST     EQU 86
jump_table
    db         7,$27,6,5,   5,6,7,$27
    db         5,4,3,3,3,   3,4,5,7,$27
    db   5,4,3,2,1,1,0,0,   0,0,1,1,2,3,4,$15

    
;Screen table:
;-------------
    
screen_table        
    REPT 6,sprite_screen_row 
        REPT 4, sprite_screen_line
            dw 16384+2048+2048+((sprite_screen_row+2)*32)+(sprite_screen_line*512)
        ENDM
    ENDM
    REPT 16
        dw 0
    ENDM 
    

;Collision table:
;----------------

collision_table
    dw  coll0+5, coll0+5, coll0+5, coll0+5, coll0+5,  coll0+5, coll0+5
    dw coll0+10,coll0+10,coll0+10,coll0+10,coll0+10
    dw coll0+15,coll0+15,coll0+15,coll0+15,coll0+15
    dw coll0+20,coll0+20,coll0+20,coll0+20,coll0+20
    dw coll0+25,coll0+25,coll0+25,coll0+25,coll0+25,  coll0+25,coll0+25,coll0+25,coll0+25,coll0+25 

    
;Drop table:
;-----------
drop_table
    db 0, 1, 2, 3, 4, 4, 4, 8, 8, 9, 9, 9, 13, 13
    db 14
    db 15, 15, 18, 18, 20, 20, 20, 24, 24, 24, 25, 26, 27, 28, 28, 28, 28        
    
    
;Road table:
;-----------

road_table
    db   0,240,238,236,  234,233,231,229,  227,226,224,222
    db 220,218,216,214,  212,210,208,205,  203,200,197,194
    
    db   0,192,190,188,187,185,183,181,  180,178,176,174,172,170,168,166
    db 164,162,160,158,156,154,152,150,  148,146,144,142,139,137,135,133
    db 130,128,125,123,120,117,115,112
    db 109,107,104,101,099,096,094,091
    db 088,085,083,080,077,075,072,070
    db 067,064,061,059,056,053,051,048
    db 045,043,041,039,037,035,033,031
    
    db 028,024,020,016    
    
    
;Road map:
;---------    
    
    PAGE_ALIGN
road_map

    ;  col1/blk      col1      col1/col2 col2/col1     col2      col3/col2 col2/col3     col3      col3/col4 col4/col3     col4      col5/col4 col4/col5     col5      col5/blk
    db 00000100b,  00100110b,  00011100b,00100011b,  00011110b,  00011100b,00100011b,  00100110b,  00011100b,00100011b,  00011110b,  00011100b,00100011b,  00100110b,  00000100b,  0
    db 01000010b,  01010010b,  01111010b,01010111b,  01111111b,  01111001b,01001111b,  01001001b,  01111001b,01001111b,  01111111b,  01111010b,01010111b,  01010010b,  01000010b,  0
    db 00000111b,  00111111b,  00001111b,00111001b,  00001001b,  00001110b,00110001b,  00110110b,  00001110b,00110001b,  00001001b,  00001111b,00111001b,  00111111b,  00000111b,  0
    db 01000010b,  01010010b,  01111010b,01010111b,  01111111b,  01111001b,01001111b,  01001001b,  01111001b,01001111b,  01111111b,  01111010b,01010111b,  01010010b,  01000010b,  0
    db 00000111b,  00111111b,  00001111b,00111001b,  00001001b,  00001000b,00000001b,  00000111b,  00001000b,00000001b,  00001001b,  00001111b,00111001b,  00111111b,  00000111b,  0
    db 01000010b,  01010010b,  01000010b,01010000b,  00000111b,  00000000b,00000000b,  00000111b,  00000000b,00000000b,  00000111b,  01000010b,01010000b,  01010010b,  01000010b,  0
    db 00000111b,  00111111b,  00001111b,00111001b,  00001001b,  00001000b,00000001b,  00000111b,  00001000b,00000001b,  00001001b,  00001111b,00111001b,  00111111b,  00000111b,  0
    db 01000010b,  01010010b,  01111010b,01010111b,  01111111b,  01111001b,01001111b,  01001001b,  01111001b,01001111b,  01111111b,  01111010b,01010111b,  01010010b,  01000010b,  0
    db 00000111b,  00111111b,  00001111b,00111001b,  00001001b,  00001110b,00110001b,  00110110b,  00001110b,00110001b,  00001001b,  00001111b,00111001b,  00111111b,  00000111b,  0
    db 01000010b,  01010010b,  01111010b,01010111b,  01111111b,  01111001b,01001111b,  01001001b,  01111001b,01001111b,  01111111b,  01111010b,01010111b,  01010010b,  01000010b,  0
    db 00000000b,  00000111b,  00001000b,00000001b,  00001001b,  00001110b,00110001b,  00110110b,  00001110b,00110001b,  00001001b,  00001000b,00000001b,  00000111b,  00000000b,  0
    db 00000000b,  00000111b,  01000000b,01000000b,  00000111b,  01000001b,01001000b,  01001001b,  01000001b,01001000b,  00000111b,  01000000b,01000000b,  00000111b,  01000000b,  0
    db 00000000b,  00000111b,  00001000b,00000001b,  00001001b,  00001110b,00110001b,  00110110b,  00001110b,00110001b,  00001001b,  00001000b,00000001b,  00000111b,  00000000b,  0
    db 01000010b,  01010010b,  01111010b,01010111b,  01111111b,  01111001b,01001111b,  01001001b,  01111001b,01001111b,  01111111b,  01111010b,01010111b,  01010010b,  01000010b,  0
    db 00000111b,  00111111b,  00001111b,00111001b,  00001001b,  00001110b,00110001b,  00110110b,  00001110b,00110001b,  00001001b,  00001111b,00111001b,  00111111b,  00000111b,  0
    db 01000010b,  01010010b,  01111010b,01010111b,  01111111b,  01111001b,01001111b,  01001001b,  01111001b,01001111b,  01111111b,  01111010b,01010111b,  01010010b,  01000010b,  0
        
    
;Colour Map
;----------

white_row
    ds 16, 00000111b

    
;Haze section, four lines interleaved with white:
colour_map_a
    db 55,23,55,23,7,  5,3,6,3,5,  7,7,23,55,23,55
    db 23,7,7,7,7,  5,3,6,3,5,  7,7,7,7,7,23
    db 55,7,55,7,7,  5,3,6,3,5,  7,7,7,55,7,55
    db 23,7,7,7,7,  5,3,6,3,5,  7,7,7,7,7,23
        
;Fade section, eight lines:
colour_map_b
    db   7,7,7,7,7,  70,65,69,65,70,  7,7,7,7,7,7
    db   7,7,7,7,7,  70,65,69,65,70,  7,7,7,7,7,7
    db   7,7,7,7,7,  70,65,69,65,70,  7,7,7,7,7,7
    db   7,7,7,7,7,  70,65,69,65,70,  7,7,7,7,7,7
    db   7,7,7,7,7,  70,65,69,65,70,  7,7,7,7,7,7
    db   7,7,7,7,7,  70,65,69,65,70,  7,7,7,7,7,7
    db   7,7,7,7,7,  70,65,69,65,70,  7,7,7,7,7,7
    db   7,7,7,7,7,  70,65,69,65,70,  7,7,7,7,7,7

;Narrow section, 8 rows less one line (which will be doubled):
colour_map_c
    REPT 6
        db    7,7,7,  01000101b,01000101b,01011101b, 01011110b,01110110b,01011110b, 01011101b,01000101b,01000101b,  7,7,7,7
        db    7,7,7,  01000101b,01000101b,01011101b, 01011110b,01110110b,01011110b, 01011101b,01000101b,01000101b,  7,7,7,7
        db    7,7,7,  01000101b,01000101b,01011101b, 01011110b,01110110b,01011110b, 01011101b,01000101b,01000101b,  7,7,7,7
        db    7,7,7,  01000101b,01000101b,01011101b, 01011110b,01110110b,01011110b, 01011101b,01000101b,01000101b,  7,7,7,7
        db    7,7,7,  01000101b,01000101b,01011101b, 01011110b,01110110b,01011110b, 01011101b,01000101b,01000101b,  7,7,7,7
        db    7,7,7,  01000101b,01000101b,01011101b, 01011110b,01110110b,01011110b, 01011101b,01000101b,01000101b,  7,7,7,7
        db    7,7,7,  01000101b,01000101b,01011101b, 01011110b,01110110b,01011110b, 01011101b,01000101b,01000101b,  7,7,7,7
        db    7,7,7,  01000101b,01000101b,01011101b, 01011110b,01110110b,01011110b, 01011101b,01000101b,01000101b,  7,7,7,7
    ENDM        
    REPT 7
        db   7,7,  01000101b,01000101b,01011101b, 01011011b, 01011110b,01110110b,01011110b, 01011011b, 01011101b,01000101b,01000101b,  7,7,7
        db   7,7,  01000101b,01000101b,01011101b, 01011011b, 01011110b,01110110b,01011110b, 01011011b, 01011101b,01000101b,01000101b,  7,7,7
    ENDM        
        db   7,7,  01000101b,01000101b,01011101b, 01011011b, 01011110b,01110110b,01011110b, 01011011b, 01011101b,01000101b,01000101b,  7,7,7
        
;Wide section, 6 rows of 2-line multicolour, last row 4 lines:
colour_map_d
    REPT 6
        db 7,7,7,7,7,7,  5,5,5,5,29,29,29, 30,30,30,30,30, 29,29,29,5,5,5,5,  7,7,7,7,7,7,7 
        db 7,7,7,7,7,7,  1,1,1,1,57,57,57, 58,58,58,58,58, 57,57,57,1,1,1,1,  7,7,7,7,7,7,7
    ENDM    
    REPT 5
        db 7,7,    5,5,5,5,5,29,29,29,29,29, 30,30,30,30,30,30,30, 29,29,29,29,29,5,5,5,5,5,    7,7,7 
        db 7,7,7,    1,1,1,1,57,57,57,57,57, 58,58,58,58,58,58,58, 57,57,57,57,57,1,1,1,1,    7,7,7,7
    ENDM
coll0 EQU $-32
        db 7,6,  5,5,5,5,  29,29,29,29,29,29, 30,30,30,30,30,30,30, 29,29,29,29,29,29,  5,5,5,5,  6,7,7 
    
    REPT 16
        REPT 32
            db 0
        ENDM
    ENDM

    
;Boost Gauge Colour Table    
    
pre_gauge_table
    ds 8,7
gauge_table
    REPT 10
        db 01010111b, 00000111b
    ENDM
    db 01101111b, 00000111b
    REPT 10
        db 01110111b, 00000111b
    ENDM
    db 01101111b, 00000111b
    REPT 10
        db 01111111b, 00000111b
    ENDM
post_gauge_table         
    ds 8,7
end_gauge_table         

gauge_colours
    db 00000111b
    REPT 10
        db 01010111b
    ENDM
    db 01101111b
    REPT 10
        db 01110111b
    ENDM
    db 01101111b
    REPT 10
        db 01111111b
    ENDM


;Sound FX Tables (80 bytes each)   

sound_tables
sound_table_0       ; Silent
    REPT 40
        db 0, DEBUG_BORDER_1
    ENDM    
    
sound_table_1       ; Jump init
    REPT 10
        db 24, 24, 24, 24,  0,  0,  0,  0
    ENDM    
    
sound_table_2       ; Jump bounce
    REPT 4
        ds 10,24
        ds 10, 0
    ENDM
    
sound_table_3       ; Switch lanes
    ds  1,24
    ds 38,0    
    ds  1,24
    ds 38,0    
    ds  1,24
    ds  1,0    

BOOST_TING_TIME EQU 8
sound_table_4       ; Boost ting
    REPT 10
        db 24, 0, 24,  0, 24, 24,  0,  0
    ENDM    

sound_table_n       ; Boost noise
    db 0, 0,24, 0, 0, 0,24, 0, 0, 0,24, 0, 0, 0, 0, 0
    db 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,24, 0, 0, 0, 0, 0
    db 0, 0, 0, 0, 0, 0,24, 0, 0, 0, 0, 0, 0, 0,24, 0
    db 0, 0,24, 0, 0, 0, 0, 0, 0, 0,24, 0, 0, 0, 0, 0
    db 0, 0, 0, 0, 0, 0,24, 0, 0, 0, 0, 0, 0, 0, 0, 0 ;80 bytes
    
    db 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,24, 0
    db 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,24, 0, 0, 0, 0, 0
    db 0, 0, 0, 0, 0, 0,24, 0, 0, 0,24, 0, 0, 0, 0, 0
    db 0, 0,24, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,24, 0 ;+64 bytes

             
;Working Road Data:
;------------------

org TOP_PAGE
road_data
    incbin "myroad.bin"
    
    
;Binary data to low memory:
;--------------------------

org ENTRY_POINT-6400
pixel_map
    incbin "rtp01_08_6400b.bin"

    
;Assembly ends:
;--------------

end ENTRY_POINT